#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <dir.h>
#include "chmlib.h"
#define PATH_MAX 512

static char *pointtoname(char *path)
{
  char *p = path + strlen(path);
  while (p > path && *(p-1) != '/')
    --p;
  return p;
}

static int makedir(char *path)
{
  char *end;
  char *p;

  p = path+3;
  while ((end = strstr(p,"/")) != NULL)
  {
    int h;
    struct ffblk f;
    *end = '\0';
    h = findfirst(path,&f,FA_DIREC);
    if (!h)
    {
      findclose(&f);
      *end = '/';
      p = end + 1;
      continue;
    }

    if (mkdir(path)!=0)
    {
      fprintf(stderr, "Couldn't make directory %s\n", path);
      perror("mkdir");
      *end = '/';
      return 0;
    }
    *end = '/';
    p = end + 1;
  }
  return 1;
}

static int savetofile(chmfile *c, char *target, int extractwithoutpath)
{
  ulong length;
  ubyte *outbuf;
  char *rtarget;
  char *p;
  FILE *f;
  char path[PATH_MAX];
  char fullpath[PATH_MAX];

  if (chm_getfile(c, target, &length, &outbuf) != 0)
    return 1;
  getcurdir(0,path);
  fullpath[0]=getdisk()+'A'; fullpath[1]=':'; fullpath[2]='/'; fullpath[3]='\0';
  strcat(fullpath,path);
  while ((p = strstr(fullpath,"\\")) != NULL)
    *p = '/';
  if (fullpath[strlen(fullpath)-1] != '/')
    strcat(fullpath,"/");
  strcat(fullpath,extractwithoutpath?pointtoname(target):target+1);
  if (!extractwithoutpath)
    if (!makedir(fullpath))
      return -1;
  if (fullpath[strlen(fullpath)-1]!='/')
  {
    f = fopen(fullpath, "wb");
    if (!f)
    {
      fprintf(stderr, "Couldn't open %s\n", fullpath);
      perror("fopen");
      return -1;
    }
    fwrite(outbuf, 1, length, f);
    fclose(f);
  }
  if (outbuf) free(outbuf);
  return 0;
}

static extractdir(chmfile *c, char *path, int extractwithoutpath)
{
  char filter[PATH_MAX];
  int i;
  strcpy(filter,path);
  strcat(filter,"/");
  if (savetofile(c,filter,extractwithoutpath)!=0)
    return -1;
  for (i = 0; i < c->dir->nentries; i++)
  {
    if (!strncmp(filter,c->dir->entry[i].name,strlen(filter)))
      if (savetofile(c,c->dir->entry[i].name,extractwithoutpath)!=0)
        return -1;
  }
  return 0;
}

static void usage(void)
{
  printf("CHMView v1.0 December 1 2002\n\n");
  printf("Usage: chmview <option> chmfile [/file_to_extract]\n");
  printf("         Options:\n");
  printf("           l - list archive contents\n");
  printf("           e - extract without path names\n");
  printf("           x - extract\n");
  printf("         Notes:\n");
  printf("           When extracting specific files, file name MUST start with `/'.\n");
  printf("           No file masks should be used.\n\n");
  printf("Author: Alex Yaroslavsky <totrtx@barak-online.net>.\n");
  printf("Using CHMLib library by Matthew T. Russotto\n");
  printf("and LZX library by Stuart Caie <kyzer@4u.net>.\n");
}

int main(int argc, char *argv[])
{
  chmfile *c;
  char *infname;
  char *command;
  int i;

  if (argc < 3)
  {
    usage();
    exit(-1);
  }

  infname = argv[2];
  c = chm_openfile(infname);

  if (!c)
  {
    perror("chm_openfile");
    exit(-1);
  }

  command = argv[1];

  switch (command[0])
  {
    case 'l':
    {
      printf("--------\n");
      for (i = 0; i < c->dir->nentries; i++)
      {
        printf("%12.12u ",c->dir->entry[i].length);
        if (c->dir->entry[i].name[0] != '/')
          printf("/");
        printf("%s\n",c->dir->entry[i].name);
      }
      printf("--------\n");
      break;
    }

    case 'e':
    case 'x':
    {
      if (argc == 4)
      {
        char *p;
        int s;
        while ((p = strstr(argv[3],"\\")) != NULL)
          *p = '/';
        s=savetofile(c,argv[3],command[0]=='e'?1:0);
        if (s==-1)
          exit(-1);
        if (s==1 && extractdir(c,argv[3],command[0]=='e'?1:0))
          exit(-1);
      }
      else
      {
        int i;
        for (i = 0; i < c->dir->nentries; i++)
        {
          int s;
          s = savetofile(c,c->dir->entry[i].name,command[0]=='e'?1:0);
          if (s==-1)
            exit(-1);
          if (s==1 && extractdir(c,c->dir->entry[i].name,command[0]=='e'?1:0))
            exit(-1);
        }
      }
      break;
    }
  }
  chm_close(c);

  return 0;
}
